import numpy as np
import pyopencl as cl

from mot.cl_routines.base import CLRoutine
from mot.load_balance_strategies import Worker
from mot.random123 import RandomStartingPoint, get_random123_cl_code

__author__ = 'Robbert Harms'
__date__ = "2014-10-29"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Random123GeneratorBase(CLRoutine):

    def __init__(self, starting_point=None, **kwargs):
        """Create the random123 basis for generating a list of random numbers.

        *From the Random123 documentation:*

        Unlike conventional RNGs, counter-based RNGs are stateless functions (or function classes i.e. functors)
        whose arguments are a counter, and a key and returns a result of the same type as the counter.

        .. code-block: c

            result = CBRNGname(counter, key)

        The returned result is a deterministic function of the key and counter, i.e. a unique (counter, key)
        tuple will always produce the same result. The result is highly sensitive to small changes in the inputs,
        so that the sequence of values produced by simply incrementing the counter (or key) is effectively
        indistinguishable from a sequence of samples of a uniformly distributed random variable.

        All the Random123 generators are counter-based RNGs that use integer multiplication, xor and permutation
        of W-bit words to scramble its N-word input key.

        Args:
            starting_point (mot.random123.Random123StartingPoint): object that provides the generator
                with the starting points (key, counter) tuple. If None is given we will use the
                ``RandomStartingPoint``.
        """
        super(Random123GeneratorBase, self).__init__(**kwargs)

        starting_point = starting_point or RandomStartingPoint()

        self.context = self.cl_environments[0].get_cl_context().context

        self._key = starting_point.get_key()
        self._counter = starting_point.get_counter()

    def generate_uniform(self, nmr_samples, minimum=0, maximum=1, dtype=None):
        """Draw random samples from the uniform distribution.

        Args:
            nmr_samples (int): The number of samples to draw
            minimum (double): The minimum value of the random numbers
            maximum (double): The minimum value of the random numbers
            dtype (np.dtype): the numpy datatype, either one of float32 (default) or float64.

        Returns:
            ndarray: A numpy array with nmr_samples random samples drawn from the uniform distribution.
        """
        dtype = dtype or np.float32
        if dtype not in (np.float32, np.float64):
            raise ValueError('The given dtype should be either float32 or float64, {} given.'.format(
                dtype.__class__.__name__))

        c_type = 'float'
        if dtype == np.float64:
            c_type = "double"

        return self._generate_samples(nmr_samples, self._get_uniform_kernel(minimum, maximum, c_type))

    def generate_gaussian(self, nmr_samples, mean=0, std=1, dtype=None):
        """Draw random samples from the Gaussian distribution.

        Args:
            nmr_samples (int): The number of samples to draw
            mean (double): The mean of the distribution
            std (double): The standard deviation or the distribution
            dtype (np.dtype): the numpy datatype, either one of float32 (default) or float64.

        Returns:
            ndarray: A numpy array with nmr_samples random samples drawn from the Gaussian distribution.
        """
        dtype = dtype or np.float32
        if dtype not in (np.float32, np.float64):
            raise ValueError('The given dtype should be either float32 or float64, {} given.'.format(
                dtype.__class__.__name__))

        c_type = 'float'
        if dtype == np.float64:
            c_type = "double"

        return self._generate_samples(nmr_samples, self._get_gaussian_kernel(mean, std, c_type))

    def _generate_samples(self, nmr_samples, kernel_source):
        padding = (-nmr_samples) % 4
        nmr_samples += padding
        samples = np.zeros((nmr_samples,), dtype=np.float32)

        workers = self._create_workers(lambda cl_environment: _Random123Worker(cl_environment, samples,
                                                                               kernel_source, self._key, self._counter))
        self.load_balancer.process(workers, nmr_samples // 4)

        if padding:
            return samples[:-padding]
        return samples

    def _get_rand123_init_cl_code(self):
        if len(self._key):
            return 'rand123_initialize_data_extra_precision_constmem(rand123_counter, rand123_key)'
        else:
            return 'rand123_initialize_data_constmem(rand123_counter)'

    def _get_uniform_kernel(self, min_val, max_val, c_type):
        src = get_random123_cl_code()
        src += '''
            __kernel void generate(constant uint* rand123_counter,
                                   ''' + ('constant uint* rand123_key,' if len(self._key) else '') + '''
                                   global ''' + c_type + '''* samples){

                rand123_data rng_data = ''' + self._get_rand123_init_cl_code() + ''';

                ''' + c_type + '''4 randomnr =  rand123_uniform_''' + c_type + '''4(&rng_data);

                int gid = get_global_id(0);

                samples[gid * 4] = ''' + str(min_val) + ''' + randomnr.x * ''' + str(max_val - min_val) + ''';
                samples[gid * 4 + 1] = ''' + str(min_val) + ''' + randomnr.y * ''' + str(max_val - min_val) + ''';
                samples[gid * 4 + 2] = ''' + str(min_val) + ''' + randomnr.z * ''' + str(max_val - min_val) + ''';
                samples[gid * 4 + 3] = ''' + str(min_val) + ''' + randomnr.w * ''' + str(max_val - min_val) + ''';
            }
        '''
        return src

    def _get_gaussian_kernel(self, mean, std, c_type):
        src = get_random123_cl_code()
        src += '''
            __kernel void generate(constant uint* rand123_counter,
                                   ''' + ('constant uint* rand123_key,' if len(self._key) else '') + '''
                                   global ''' + c_type + '''* samples){

                rand123_data rng_data = ''' + self._get_rand123_init_cl_code() + ''';

                ''' + c_type + '''4 randomnr =  rand123_normal_''' + c_type + '''4(&rng_data);

                int gid = get_global_id(0);

                samples[gid * 4] = ''' + str(mean) + ''' + randomnr.x * ''' + str(std) + ''';
                samples[gid * 4 + 1] = ''' + str(mean) + ''' + randomnr.y * ''' + str(std) + ''';
                samples[gid * 4 + 2] = ''' + str(mean) + ''' + randomnr.z * ''' + str(std) + ''';
                samples[gid * 4 + 3] = ''' + str(mean) + ''' + randomnr.w * ''' + str(std) + ''';
            }
        '''
        return src


class _Random123Worker(Worker):

    def __init__(self, cl_environment, samples, kernel_source, key, counter):
        super(_Random123Worker, self).__init__(cl_environment)
        self._samples = samples
        self._nmr_samples = self._samples.shape[0]
        self._kernel_source = kernel_source
        self._key = key
        self._counter = counter

        self._samples_buf = cl.Buffer(self._cl_run_context.context, cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR,
                                      hostbuf=self._samples)

        self._counter_buffer = cl.Buffer(self._cl_run_context.context,
                                         cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf=self._counter)

        if len(self._key):
            self._key_buffer = cl.Buffer(self._cl_run_context.context,
                                         cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf=self._key)

        self._kernel = self._build_kernel()

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start

        kernel_args = [self._counter_buffer]
        if len(self._key):
            kernel_args.append(self._key_buffer)
        kernel_args.append(self._samples_buf)

        event = self._kernel.generate(self._cl_run_context.queue, (int(nmr_problems), ), None,
                                      *kernel_args, global_offset=(range_start,))
        return [self._enqueue_readout(self._samples_buf, self._samples, range_start * 4, range_end * 4, [event])]

    def _get_kernel_source(self):
        return self._kernel_source



import os
from pkg_resources import resource_filename
from ...cl_routines.optimizing.base import AbstractParallelOptimizer, AbstractParallelOptimizerWorker

__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class NMSimplex(AbstractParallelOptimizer):

    default_patience = 200

    def __init__(self, patience=None, optimizer_settings=None, **kwargs):
        """Use the Nelder-Mead simplex method to calculate the optimimum.

        Args:
            patience (int):
                Used to set the maximum number of iterations to patience*(number_of_parameters+1)
            optimizer_settings (dict): the optimization settings, you can use the following:
                - scale (double): the scale of the initial simplex, default 1.0
                - alpha (double): reflection coefficient, default 1.0
                - beta (double): contraction coefficient, default 0.5
                - gamma (double); expansion coefficient, default 2.0
        """
        patience = patience or self.default_patience

        optimizer_settings = optimizer_settings or {}
        option_defaults = {'alpha': 1.0, 'beta': 0.5, 'gamma': 2.0, 'delta': 0.5, 'scale': 1.0}

        def get_value(option_name, default):
            value = optimizer_settings.get(option_name, None)
            if value is None:
                return default
            return value

        for option, default in option_defaults.items():
            optimizer_settings.update({option: get_value(option, default)})

        super(NMSimplex, self).__init__(patience=patience, optimizer_settings=optimizer_settings, **kwargs)

    def _get_worker_generator(self, *args):
        return lambda cl_environment: NMSimplexWorker(cl_environment, *args)


class NMSimplexWorker(AbstractParallelOptimizerWorker):

    def _get_optimization_function(self):
        params = {'NMR_PARAMS': self._nmr_params, 'PATIENCE': self._parent_optimizer.patience}

        for option, value in self._optimizer_settings.items():
            if option == 'scale':
                params['INITIAL_SIMPLEX_SCALES'] = '{' + ', '.join([str(value)] * self._nmr_params) + '}'
            else:
                params.update({option.upper(): value})

        body = open(os.path.abspath(resource_filename('mot', 'data/opencl/nmsimplex.pcl')), 'r').read()
        if params:
            body = body % params
        return body

    def _get_optimizer_call_name(self):
        return 'nmsimplex'



import logging
from .__version__ import VERSION, VERSION_STATUS, __version__

__author__ = 'Robbert Harms'
__date__ = '2015-01-01'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"

try:
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass
logging.getLogger(__name__).addHandler(NullHandler())


def smart_device_selection():
    """Get a list of device environments that is suitable for use in MOT.

    Returns:
        list of CLEnvironment: List with the CL device environments.
    """
    from mot.cl_environments import CLEnvironmentFactory
    return CLEnvironmentFactory.smart_device_selection()



"""The interfaces needed for models.

Since a lot of information about a model is needed to be able to optimize or sample it we encapsulate all that
information in an interface. Only objects that successful implement the interfaces in this module can be optimized or
sampled using one of the optimization or sampling routines in MOT.

These interfaces expose data and modelling code. The data is encapsulated in :class:`~mot.data_adapters.DataAdapter`
and the code should be returned as CL strings. In the future, instead of CL strings we may require returning an
encapsulating object such that we can run the computations on multiple types of runtime environments
(CUDA, plain C, ...).
"""

__author__ = 'Robbert Harms'
__date__ = "2014-03-14"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class OptimizeModelInterface(object):

    @property
    def name(self):
        """Get the name of this model. This should be overwritten by the implementing model.

        Returns:
            str: A string with the name of this model.
        """
        return ""

    @property
    def double_precision(self):
        """Flag to signal if we should use the double float type during calculations.

        By default we ask the cl routines to use the single precision float type, you can overwrite this with
        your own flags.

        Returns:
            boolean: if we would like to use double precision floating point during the calculations
        """
        return False

    def get_problems_var_data(self):
        """Get a dict with all the data per problem.

        As an example, suppose per problem instance we have the data named 'observations'.
        We should then return something like:

        .. code-block:: python

            {'observations': SimpleDataAdapter(ndarray((<nmr_problems>, <number of items>)), ...), ...}

        Returns:
            dict: A dictionary where each element holds a SimpleDataAdapter from which to get the data
                The data in the adapter should contain for each problem (first dimension / rows) a
                number of items (second dimension / columns) that is used in the evaluation function.
        """
        raise NotImplementedError

    def get_problems_protocol_data(self):
        """Get a dict with the data that is constant for each problem, but differs per measurement.

        Returns:
            dict: A dict where each element holds a SimpleDataAdapter that is used in the evaluation function.
                In the CL kernel, each key is used as name in the cl data struct.
        """
        raise NotImplementedError

    def get_model_data(self):
        """Get a dict with all the model data. This is model data necessary for computations in the model.

        For data that is to large to inline in the kernel this data type may be useful.

        Returns:
            dict: The dict with all the model data in DataAdapters.
                The names of the keys are used in the CL cl data structs.
        """

    def get_nmr_problems(self):
        """Get the number of problems we need to analyze.

        Returns:
            int: A single integer specifying the number of problem instances
        """
        raise NotImplementedError

    def get_model_eval_function(self, func_name='evaluateModel'):
        """Get the evaluation function that evaluates the model at the given parameters.

        This returned function should not do any error calculations,
        it should merely return the result of evaluating the model for the given parameters.

        Please make sure the sign of the return value is correct given the following. The minimization
        routines may make use of this function and get_observation_return_function to build their
        own objective function. This is always done as: observation() - evaluation(). This means that if you
        want to optimize a function without observation data you need to make sure the evaluation function returns
        the answers with the right sign.

        Args:
            func_name (string): specifies the name of the function.

        Returns:
            str: An CL function with the signature:
                .. code-block:: c

                    mot_float_type <func_name>(const optimize_data* const data, const mot_float_type* const x,
                                            const int observation_index);
        """
        raise NotImplementedError

    def get_observation_return_function(self, func_name='getObservation'):
        """Get the CL function that returns the observation for the given problem.

        Args:
            func_name (string): specifies the name of the function.

        Returns:
            str: An CL function with the signature:
                .. code-block:: c

                    mot_float_type <func_name>(const optimize_data* const data, const int observation_index);
        """
        raise NotImplementedError

    def get_objective_function(self, func_name="calculateObjective"):
        """Get the objective function that evaluates the entire problem instance under a noise model.

        Args:
            func_name (string): specifies the name of the function.

        Returns:
            str: A CL function with signature:
                .. code-block:: c

                    mot_float_type <func_name>(const optimize_data* const data, mot_float_type* const x);
        """
        raise NotImplementedError

    def get_objective_list_function(self, func_name="calculateObjectiveList"):
        """Get the objective function returning a list with the objective functions per instance point.

        This function is used by some evaluation routines (like for example LevenbergMarquardt) that need
        a list of objective values (one per instance point), instead of a single objective function scalar.

        Args:
            func_name (str): the name of the function

        Returns:
            str: A CL function with signature:
                .. code-block:: c

                    mot_float_type <func_name>(const optimize_data* const data, mot_float_type* const x,
                                               mot_float_type* result);
        """
        raise NotImplementedError

    def get_initial_parameters(self, results_dict=None):
        """Get a two dimensional matrix with the initial parameters (starting points) for every voxel.

        Optionally, one may specify a list of previously calculated results which may be applicable to the model.
        If a parameter is found in the results_dict, those values are used for the initial parameters.

        Args:
            results_dict (dict): a dictionary with for every parameter name, a value per voxel which is (for example)
                the result of a previous calculation.

        Returns:
            ndarray: A two dimensional matrix with on the first axis the problem instances and on the second
                the parameter values per problem instance
        """
        raise NotImplementedError

    def get_lower_bounds(self):
        """Get for each estimable parameter the lower bounds.

        Returns:
            ndarray: An numpy row with on each column a single value, the lower bound for that parameter. This value
            can be the literal string '-inf' for infinity.
        """
        raise NotImplementedError

    def get_upper_bounds(self):
        """Get for each estimable parameter the upper bounds.

        Returns:
            ndarray: An numpy row with on each column a single value, the upper bound for that parameter. This value
            can be the literal string 'inf' for infinity.
        """
        raise NotImplementedError

    def get_optimized_param_names(self):
        """Get a list of names with the free parameter names (the parameters that are estimated by the routines).

        The function get_optimization_output_param_names() returns the names of all the parameter names,
        including fixed and static parameters. This should only return the names of the parameters
        that are actually used in the optimization.

        Returns:
            list of str: A list with the parameter names (in dot format) of all the estimated (free) parameters.
        """
        raise NotImplementedError

    def get_optimization_output_param_names(self):
        """Get a list with the names of the parameters, this is the list of keys to the titles and results.

        See get_optimized_param_names() for getting the names of the parameters that are actually being optimized.

        This should be a complete overview of all the maps returned from optimizing this model.

        Returns:
            list of str: a list with the parameter names
        """
        raise NotImplementedError

    def get_nmr_inst_per_problem(self):
        """Get the number of instances/data points per problem.

        The minimum is one instance per problem.

        This number represents the number of data points

        Returns:
            int: the number of instances per problem.
        """
        raise NotImplementedError

    def get_nmr_estimable_parameters(self):
        """Get the number of estimable parameters.

        Returns:
            int: the number of estimable parameters
        """
        raise NotImplementedError

    def get_parameter_codec(self):
        """Get the parameter codec from this model. This should be an implementation of AbstractCodec.

        Returns:
            AbstractCodec or None: An abstract codec model that holds the CL code for the codec transformations.
            This function may also return None, which indicates that no parameter codec is supposed to be used.
        """
        raise NotImplementedError

    def get_final_parameter_transformations(self, func_name='applyFinalParameterTransformations'):
        """Get the transformations that must be applied at the end of an optimization (or sampling) routine.

        These transformations must contain all parameter dependencies, that is, all transformation happening in the
        model function which do not happen in the codec must be present in this function.

        Suppose an optimization routine finds a set of parameters X to the the optimal set of parameters. In the
        evaluation function this set of parameters might have been transformed to a new set of parameters X' by the
        parameter dependencies. Since we, in the end, are interested in the set of parameters X', we have to apply
        the exact same transformations at the end of the optimization routine as happened in the evaluation function.

        Args:
            func_name (string): specifies the name of the function.

        Returns:
            str or None: Return None if this function is not used, else a function of the kind:
                .. code-block:: c

                    void <func_name>(const optimize_data* data, mot_float_type* x);

            Which is called for every voxel and must in place edit the x variable.
        """
        raise NotImplementedError

    def finalize_optimization_results(self, results_dict):
        """After optimization create the final dictionary with the result maps.

        In this location extra maps can be added to the results dictionary.

        This function behaves as a procedure and as a function. The input dict can be updated in place, but it should
        also return a dict but that is merely for the purpose of chaining.

        Args:
            results_dict (dict): A dictionary with as keys the names of the parameters and as values the 1d maps with
                for each voxel the optimized parameter value. The given dictionary can be altered by this function.

        Returns:
            dict: The same result dictionary but with updated values or with additional maps.
                It should at least return the results_dict.
        """
        return results_dict


class SampleModelInterface(OptimizeModelInterface):

    def __init__(self):
        """Extends the OptimizeModelInterface with information for sampling purposes.

        To be able to sample a model we need to have a:

        * log likelihood function
        * a proposal function
        * a prior function


        Proposal functions can be symmetric (if it holds that ``q(x|x') == q(x'|x)``) or
        non symmetric (i.e. ``q(x|x') != q(x'|x)``). In the case of non-symmetric proposals we need to
        have a function to get the probability log likelihood of the proposal.
        This indicates the need for two more pieces of information:

        * test if the proposal is symmetric
        * proposal log PDF function


        A trick in sampling is to have auto-adapting proposals. These proposals commonly have a distribution with a
        standard deviation that varies in time. The idea is that if the distribution is too tight (low std) only a few
        of the proposed samples are accepted and we need to broaden the distribution (increase the std). On the other
        hand, if the std is to high we are jumping around to much in search space. This leads us to the following
        additional functionality:

        * proposal state update function


        Since OpenCL < 2.1 does not allow for state variables in functions and does not have classes we need to find
        a way to store the state of the proposal distribution inside the kernel function. For that, each proposal
        CL function has as additional parameter the ``proposal_state``. The initial state can be obtained from
        this class and needs to be handed to the proposal functions.
        """
        super(SampleModelInterface, self).__init__()

    def get_proposal_state(self):
        """Get a list of parameter values for the adaptable proposal parameters.

        Returns:
            list: list of float with the proposal parameter values that are adaptable.
        """
        raise NotImplementedError

    def get_log_likelihood_function(self, func_name="getLogLikelihood", evaluation_model=None, full_likelihood=True):
        """Get the CL Log Likelihood function that evaluates the entire problem instance under a noise model

        Args:
            func_name (string): specifies the name of the function.
            evaluation_model (EvaluationModel): the evaluation model to use for the log likelihood. If not given
                we use the one defined in the model.
            full_likelihood (boolean): if we want the complete likelihood, or if we can drop the constant terms.
                The default is the complete likelihood. Disable for speed.

        Returns:
            str: A function of the kind:
                .. code-block:: c

                    mot_float_type <func_name>(const optimize_data* const data, mot_float_type* const x);
        """
        raise NotImplementedError

    def is_proposal_symmetric(self):
        """Check if the entire proposal distribution is symmetric: ``q(x|x') == q(x'|x)``.

        Returns:
            boolean: True if the proposal distribution is symmetric, false otherwise.
        """
        raise NotImplementedError

    def get_proposal_logpdf(self, func_name='getProposalLogPDF'):
        """Get the probability density function of the proposal in log space (as a CL string).

        This density function is used if the proposal is not symmetric.

        Args:
            func_name (str): the CL function name of the returned function

        Returns:
            str: A function with the signature:
                .. code-block:: c

                    mot_float_type <func_name>(const int i, const mot_float_type proposal,
                                               const mot_float_type current, mot_float_type* const proposal_state)

            Where ``i`` is the index of the parameter we would like to get the proposal from, ``current`` is the current
            value of that parameter and ``proposal`` the proposal value of the parameter. The final argument
            ``proposal_state`` are the current settings of the proposal function.

            It should return for the requested parameter a value ``q(proposal | current)``, the log Probability
            Density Function (log PDF) of the proposal given the current value.
        """
        raise NotImplementedError

    def get_proposal_function(self, func_name='getProposal'):
        """Get a proposal function that returns proposals for a requested parameter.

        Args:
            func_name (str): the CL function name of the returned function

        Returns:
            str: A function with the signature:
                .. code-block:: c

                    mot_float_type <func_name>(const int i, const mot_float_type current, void* rng_data,
                                               mot_float_type* const proposal_state)

            Where ``i`` is the index of the parameter for which we want the proposal and ``current`` is the current
            value of that parameter. The argument ``proposal_state`` is the state of the proposal distribution.
            One can obtain random numbers with:
            .. code-block:: c

                float randomnr = frand(rng_data);
        """
        raise NotImplementedError

    def get_proposal_state_update_function(self, func_name='updateProposalState'):
        """Get a function that can update the parameters of the proposals

        Args:
            func_name (str): the CL function name of the returned function

        Returns:
            str: A function with the signature:
                .. code-block:: c

                    void <func_name>(uint* const ac_between_proposal_updates, const uint proposal_update_intervals,
                                     mot_float_type* const proposal_state);

            Where ``ac_between_proposal_updates`` is the acceptance count in between proposal updates,
            ``proposal_update_intervals`` is the interval at which we update the proposals and ``proposal_state``
            is the current list of proposal parameters (to be updated in place). Please note that
            ``ac_between_proposal_updates`` is per sampled parameter while ``proposal_state`` is
            per proposal parameter. If the number of parameters of a proposal function is not 1, than the
            two arrays do not share the same indices.
        """
        raise NotImplementedError

    def get_log_prior_function(self, func_name='getLogPrior'):
        """Get the prior function that returns a double representing the prior information about the given parameters.

        The prior function must be in log space.

        Args:
            func_name (str): the CL function name of the returned function

        Returns:
            str: A function with the signature:
                .. code-block:: c

                    mot_float_type <func_name>(const mot_float_type* const x);

            Which is called by the sampling routine to calculate the posterior probability.
        """
        raise NotImplementedError

    def samples_to_statistics(self, samples_dict):
        """Create statistics out of the given set of samples (in a dictionary).

        Args:
            samples_dict (dict): Keys being the parameter names, values the roi list in 2d
                (1st dim. is voxel, 2nd dim. is samples).

        Returns:
            dict: The same dictionary but with statistical maps (mean, avg etc.) for each parameter, instead of the raw
            samples. In essence this is where one can place the logic to go from samples to meaningful maps.
        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# CAVEAT UTILITOR
#
# This file was automatically generated by Grako.
#
#    https://pypi.python.org/pypi/grako/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.


from __future__ import print_function, division, absolute_import, unicode_literals

from grako.parsing import graken, Parser
from grako.util import re, RE_FLAGS


__version__ = (2016, 1, 5, 17, 39, 0, 1)

__all__ = [
    'CLDataTypeParser',
    'CLDataTypeSemantics',
    'main'
]


class CLDataTypeParser(Parser):
    def __init__(self,
                 whitespace=None,
                 nameguard=None,
                 comments_re=None,
                 eol_comments_re=None,
                 ignorecase=None,
                 left_recursion=True,
                 **kwargs):
        super(CLDataTypeParser, self).__init__(
            whitespace=whitespace,
            nameguard=nameguard,
            comments_re=comments_re,
            eol_comments_re=eol_comments_re,
            ignorecase=ignorecase,
            left_recursion=left_recursion,
            **kwargs
        )

    @graken()
    def _result_(self):
        self._expr_()

    @graken()
    def _expr_(self):
        with self._optional():
            self._address_space_qualifier_()

        def block0():
            self._pre_data_type_type_qualifiers_()
        self._closure(block0)
        self._data_type_()
        with self._optional():
            self._is_pointer_()
            with self._optional():
                self._post_data_type_type_qualifier_()

    @graken()
    def _data_type_(self):
        with self._choice():
            with self._option():
                self._scalar_data_type_()
            with self._option():
                self._vector_data_type_()
            with self._option():
                self._user_data_type_()
            self._error('no available options')

    @graken()
    def _is_pointer_(self):
        self._token('*')

    @graken()
    def _scalar_data_type_(self):
        with self._choice():
            with self._option():
                self._token('bool')
            with self._option():
                self._token('char')
            with self._option():
                self._token('cl_char')
            with self._option():
                self._token('unsigned char')
            with self._option():
                self._token('uchar')
            with self._option():
                self._token('cl_uchar')
            with self._option():
                self._token('short')
            with self._option():
                self._token('cl_short')
            with self._option():
                self._token('unsigned short')
            with self._option():
                self._token('ushort')
            with self._option():
                self._token('int')
            with self._option():
                self._token('unsigned int')
            with self._option():
                self._token('uint')
            with self._option():
                self._token('long')
            with self._option():
                self._token('unsigned long')
            with self._option():
                self._token('ulong')
            with self._option():
                self._token('float')
            with self._option():
                self._token('half')
            with self._option():
                self._token('size_t')
            with self._option():
                self._token('ptrdiff_t')
            with self._option():
                self._token('intptr_t')
            with self._option():
                self._token('uintptr_t')
            with self._option():
                self._token('void')
            with self._option():
                self._token('double')
            with self._option():
                self._token('half')
            self._error('expecting one of: bool char cl_char cl_short cl_uchar double float half int intptr_t long ptrdiff_t short size_t uchar uint uintptr_t ulong unsigned char unsigned int unsigned long unsigned short ushort void')

    @graken()
    def _vector_data_type_(self):
        self._pattern(r'(char|uchar|short|ushort|int|uint|long|ulong|float|double|half|mot_float_type)\d+')

    @graken()
    def _user_data_type_(self):
        self._pattern(r'\w+')

    @graken()
    def _address_space_qualifier_(self):
        with self._choice():
            with self._option():
                self._token('__local')
            with self._option():
                self._token('local')
            with self._option():
                self._token('__global')
            with self._option():
                self._token('global')
            with self._option():
                self._token('__constant')
            with self._option():
                self._token('constant')
            with self._option():
                self._token('__private')
            with self._option():
                self._token('private')
            self._error('expecting one of: __constant __global __local __private constant global local private')

    @graken()
    def _pre_data_type_type_qualifiers_(self):
        with self._choice():
            with self._option():
                self._token('const')
            with self._option():
                self._token('restrict')
            with self._option():
                self._token('volatile')
            self._error('expecting one of: const restrict volatile')

    @graken()
    def _post_data_type_type_qualifier_(self):
        self._token('const')


class CLDataTypeSemantics(object):
    def result(self, ast):
        return ast

    def expr(self, ast):
        return ast

    def data_type(self, ast):
        return ast

    def is_pointer(self, ast):
        return ast

    def scalar_data_type(self, ast):
        return ast

    def vector_data_type(self, ast):
        return ast

    def user_data_type(self, ast):
        return ast

    def address_space_qualifier(self, ast):
        return ast

    def pre_data_type_type_qualifiers(self, ast):
        return ast

    def post_data_type_type_qualifier(self, ast):
        return ast


def main(filename, startrule, trace=False, whitespace=None, nameguard=None):
    import json
    with open(filename) as f:
        text = f.read()
    parser = CLDataTypeParser(parseinfo=False)
    ast = parser.parse(
        text,
        startrule,
        filename=filename,
        trace=trace,
        whitespace=whitespace,
        nameguard=nameguard)
    print('AST:')
    print(ast)
    print()
    print('JSON:')
    print(json.dumps(ast, indent=2))
    print()

if __name__ == '__main__':
    import argparse
    import string
    import sys

    class ListRules(argparse.Action):
        def __call__(self, parser, namespace, values, option_string):
            print('Rules:')
            for r in CLDataTypeParser.rule_list():
                print(r)
            print()
            sys.exit(0)

    parser = argparse.ArgumentParser(description="Simple parser for CLDataType.")
    parser.add_argument('-l', '--list', action=ListRules, nargs=0,
                        help="list all rules and exit")
    parser.add_argument('-n', '--no-nameguard', action='store_true',
                        dest='no_nameguard',
                        help="disable the 'nameguard' feature")
    parser.add_argument('-t', '--trace', action='store_true',
                        help="output trace information")
    parser.add_argument('-w', '--whitespace', type=str, default=string.whitespace,
                        help="whitespace specification")
    parser.add_argument('file', metavar="FILE", help="the input file to parse")
    parser.add_argument('startrule', metavar="STARTRULE",
                        help="the start rule for parsing")
    args = parser.parse_args()

    main(
        args.file,
        args.startrule,
        trace=args.trace,
        whitespace=args.whitespace,
        nameguard=not args.no_nameguard
    )



import pyopencl as cl
import numpy as np
from ...utils import ParameterCLCodeGenerator, get_float_type_def
from ...cl_routines.base import CLRoutine
from ...load_balance_strategies import Worker


__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ResidualCalculator(CLRoutine):

    def __init__(self, cl_environments=None, load_balancer=None):
        """Calculate the residuals, that is the errors, per problem instance per data point."""
        super(ResidualCalculator, self).__init__(cl_environments=cl_environments, load_balancer=load_balancer)

    def calculate(self, model, parameters_dict, model_estimates=None):
        """Calculate and return the residuals.

        Args:
            model (AbstractModel): The model to calculate the residuals of.
            parameters_dict (dict): The parameters to use in the evaluation of the model
            model_estimates (ndarray): The model estimates of the model if available. If not given we
                calculate the model estimates using the initial parameters from the model.

        Returns:
            Return per voxel the errors (eval - data) per protocol item
        """
        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64

        nmr_inst_per_problem = model.get_nmr_inst_per_problem()
        nmr_problems = model.get_nmr_problems()

        residuals = np.zeros((nmr_problems, nmr_inst_per_problem), dtype=np_dtype, order='C')
        parameters = np.require(model.get_initial_parameters(parameters_dict), np_dtype, requirements=['C', 'A', 'O'])

        if model_estimates is not None:
            model_estimates = np.require(model_estimates, np_dtype, requirements=['C', 'A', 'O'])

        workers = self._create_workers(lambda cl_environment: _ResidualCalculatorWorker(
            cl_environment, self.get_compile_flags_list(), model, parameters, residuals, model_estimates))
        self.load_balancer.process(workers, model.get_nmr_problems())

        return residuals


class _ResidualCalculatorWorker(Worker):

    def __init__(self, cl_environment, compile_flags, model, parameters, residuals, model_estimates=None):
        super(_ResidualCalculatorWorker, self).__init__(cl_environment)

        self._model = model
        self._double_precision = model.double_precision
        self._residuals = residuals
        self._parameters = parameters
        self._model_estimates = model_estimates

        self._var_data_dict = model.get_problems_var_data()
        self._protocol_data_dict = model.get_problems_protocol_data()
        self._model_data_dict = model.get_model_data()

        self._all_buffers, self._residuals_buffer = self._create_buffers()
        self._kernel = self._build_kernel(compile_flags)

    def __del__(self):
        list(buffer.release() for buffer in self._all_buffers)

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start
        event = self._kernel.get_errors(self._cl_run_context.queue, (int(nmr_problems), ), None, *self._all_buffers,
                                        global_offset=(int(range_start),))
        return [self._enqueue_readout(self._residuals_buffer, self._residuals, range_start, range_end, [event])]

    def _create_buffers(self):
        errors_buffer = cl.Buffer(self._cl_run_context.context,
                                  cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR,
                                  hostbuf=self._residuals)

        if self._model_estimates is None:
            all_buffers = [cl.Buffer(self._cl_run_context.context,
                                     cl.mem_flags.READ_ONLY | cl.mem_flags.USE_HOST_PTR,
                                     hostbuf=self._parameters),
                           errors_buffer]
        else:
            all_buffers = [cl.Buffer(self._cl_run_context.context,
                                     cl.mem_flags.READ_ONLY | cl.mem_flags.USE_HOST_PTR,
                                     hostbuf=self._model_estimates),
                           errors_buffer]

        for data in self._var_data_dict.values():
            all_buffers.append(cl.Buffer(self._cl_run_context.context,
                                         cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                                         hostbuf=data.get_opencl_data()))

        constant_buffers = self._generate_constant_buffers(self._protocol_data_dict, self._model_data_dict)
        all_buffers.extend(constant_buffers)

        return all_buffers, errors_buffer

    def _get_kernel_source(self):
        nmr_inst_per_problem = self._model.get_nmr_inst_per_problem()
        nmr_params = self._parameters.shape[1]

        param_code_gen = ParameterCLCodeGenerator(self._cl_environment.device, self._var_data_dict,
                                                  self._protocol_data_dict, self._model_data_dict)

        if self._model_estimates is None:
            kernel_param_names = ['global mot_float_type* params', 'global mot_float_type* errors']
        else:
            kernel_param_names = ['global mot_float_type* model_estimates', 'global mot_float_type* errors']

        kernel_param_names.extend(param_code_gen.get_kernel_param_names())

        kernel_source = '''
            #define NMR_INST_PER_PROBLEM ''' + str(nmr_inst_per_problem) + '''
        '''

        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += param_code_gen.get_data_struct()
        kernel_source += self._model.get_observation_return_function('getObservation')
        if self._model_estimates is None:
            kernel_source += self._model.get_model_eval_function('evaluateModel')
            kernel_source += '''
                __kernel void get_errors(
                    ''' + ",\n".join(kernel_param_names) + '''
                    ){
                        int gid = get_global_id(0);
                        ''' + param_code_gen.get_data_struct_init_assignment('data') + '''

                        mot_float_type x[''' + str(nmr_params) + '''];
                        for(int i = 0; i < ''' + str(nmr_params) + '''; i++){
                            x[i] = params[gid * ''' + str(nmr_params) + ''' + i];
                        }

                        global mot_float_type* result = errors + gid * NMR_INST_PER_PROBLEM;

                        for(int i = 0; i < NMR_INST_PER_PROBLEM; i++){
                            result[i] = getObservation(&data, i) - evaluateModel(&data, x, i);
                        }
                }
            '''
        else:
            kernel_source += '''
                __kernel void get_errors(
                    ''' + ",\n".join(kernel_param_names) + '''
                    ){
                        int gid = get_global_id(0);
                        ''' + param_code_gen.get_data_struct_init_assignment('data') + '''

                        global mot_float_type* result = errors + gid * NMR_INST_PER_PROBLEM;

                        for(int i = 0; i < NMR_INST_PER_PROBLEM; i++){
                            result[i] = getObservation(&data, i) - model_estimates[i + gid * NMR_INST_PER_PROBLEM];
                        }
                }
            '''
        return kernel_source



from functools import reduce

import numpy as np
import pyopencl as cl

__author__ = 'Robbert Harms'
__date__ = "2014-05-13"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def device_type_from_string(cl_device_type_str):
    """Converts values like ``gpu`` to a pyopencl device type string.

    Supported values are: ``accelerator``, ``cpu``, ``custom``, ``gpu``. If ``all`` is given, None is returned.

    Args:
        cl_device_type_str (str): The string we want to convert to a device type.

    Returns:
        cl.device_type: the pyopencl device type.
    """
    cl_device_type_str = cl_device_type_str.upper()
    if cl_device_type_str == 'GPU':
        return cl.device_type.GPU
    if cl_device_type_str == 'CPU':
        return cl.device_type.CPU
    if cl_device_type_str == 'ACCELERATOR':
        return cl.device_type.ACCELERATOR
    if cl_device_type_str == 'CUSTOM':
        return cl.device_type.CUSTOM
    return None


def device_supports_double(cl_device):
    """Check if the given CL device supports double

    Args:
        cl_device (pyopencl cl device): The device to check if it supports double.

    Returns:
        boolean: True if the given cl_device supports double, false otherwise.
    """
    return cl_device.get_info(cl.device_info.DOUBLE_FP_CONFIG) == 63


def results_to_dict(results, param_names):
    """Create a dictionary out of the results, which the optimizer can output to the user

    Args:
        results: a 2d (from optimization) or 3d (from sampling) array that needs to be converted to a dictionary.
        param_names (list of str): the names of the parameters, one per column

    Returns:
        dict: the results packed in a dictionary
    """
    results_slice = [slice(None)] * len(results.shape)
    d = {}
    for i in range(len(param_names)):
        results_slice[1] = i
        d[param_names[i]] = results[results_slice]
    return d


def get_float_type_def(double_precision):
    """Get the model floating point type definition.

    The MOT_INT_CMP_TYPE is meant for the select() function where you need a long in the case of double precision.

    Args:
        double_precision (boolean): if True we will use the double type for the mot_float_type type.
            Else, we will use the single precision float type for the mot_float_type type.

    Returns:
        str: defines the mot_float_type types, the epsilon and the MIN and MAX values.
    """
    if double_precision:
        return '''
            #if __OPENCL_VERSION__ <= CL_VERSION_1_1
                #pragma OPENCL EXTENSION cl_khr_fp64 : enable
            #endif

            #define mot_float_type double
            #define mot_float_type2 double2
            #define mot_float_type4 double4
            #define mot_float_type8 double8
            #define mot_float_type16 double16
            #define MOT_EPSILON DBL_EPSILON
            #define MOT_MIN DBL_MIN
            #define MOT_MAX DBL_MAX
            #define MOT_INT_CMP_TYPE long
        '''
    else:
        return '''
            #if __OPENCL_VERSION__ <= CL_VERSION_1_1
                #pragma OPENCL EXTENSION cl_khr_fp64 : enable
            #endif

            #define mot_float_type float
            #define mot_float_type2 float2
            #define mot_float_type4 float4
            #define mot_float_type8 float8
            #define mot_float_type16 float16
            #define MOT_EPSILON FLT_EPSILON
            #define MOT_MIN FLT_MIN
            #define MOT_MAX FLT_MAX
            #define MOT_INT_CMP_TYPE int
        '''


class TopologicalSort(object):

    def __init__(self, data):
        """Topological sort the the given data. The data should consist of a dictionary structure.

        Args:
            data (dict); dictionary structure where the value is a list of dependencies for that given key.
                as an example ``{'a', (), 'b': ('a',)}``, where ``a`` depends on nothing and ``b`` depends on ``a``.
        """
        self.data = data

    def get_sorted(self):
        if not len(self.data):
            return

        new_data = {}
        for k, v in self.data.items():
            new_v = set([])
            if v:
                new_v = set([e for e in v if e is not k])
            new_data.update({k: new_v})
        data = new_data

        # Find all items that don't depend on anything.
        extra_items_in_deps = reduce(set.union, data.values()) - set(data.keys())

        # Add empty dependences where needed.
        data.update({item: set() for item in extra_items_in_deps})
        while True:
            ordered = set(item for item, dep in data.items() if len(dep) == 0)
            if not ordered:
                break
            yield ordered
            data = {item: (dep - ordered) for item, dep in data.items() if item not in ordered}

        if len(data) != 0:
            raise ValueError('Cyclic dependencies exist among these items: {}'.format(', '.join(repr(x)
                                                                                                for x in data.items())))

    def get_flattened(self, sort=True):
        """Returns a single flattened list of dependencies.

        Optionally we can sort the list to make the results repeatable.

        Args:
            sort (boolean): if we want to sort the results list

        Returns:
            list: the list of dependencies in constructure order, optionally sorted
        """
        result = []
        for d in self.get_sorted():
            result.extend((sorted if sort else list)(d))
        return result


class ParameterCLCodeGenerator(object):

    def __init__(self, device, var_data_dict, protocol_data_dict, model_data_dict):
        """Generate the CL code for all the parameters in the given dictionaries.

        The dictionaries are supposed to contain DataAdapters.

        Args:
            device: the CL device we want to compile the code for
            var_data_dict (dict[str, CLDataAdapter]): the dictionary with the variable data. That is, the data
                that is different for every problem (but constant over the measurements).
            protocol_data_dict (dict[str, CLDataAdapter]): the dictionary with the protocol data. That is, the data
                that is the same for every problem, but differs per measurement.
            model_data_dict (dict[str, CLDataAdapter]): the dictionary with the model data. That is, the data
                that is the same for every problem and every measurement.
        """
        self._device = device
        self._max_constant_buffer_size = device.get_info(cl.device_info.MAX_CONSTANT_BUFFER_SIZE)
        self._max_constant_args = device.get_info(cl.device_info.MAX_CONSTANT_ARGS)
        self._var_data_dict = var_data_dict
        self._protocol_data_dict = protocol_data_dict
        self._model_data_dict = model_data_dict
        self._kernel_items = self._get_all_kernel_source_items()

    def get_data_struct(self):
        return self._kernel_items['data_struct']

    def get_kernel_param_names(self):
        return self._kernel_items['kernel_param_names']

    def get_data_struct_init_assignment(self, variable_name):
        struct_code = '0'
        if self._kernel_items['data_struct_init']:
            struct_code = ', '.join(self._kernel_items['data_struct_init'])
        return 'optimize_data ' + variable_name + ' = {' + struct_code + '};'

    def _get_all_kernel_source_items(self):
        """Get the CL strings for the kernel source items for most common CL kernels in this library."""
        constant_args_counter = 0

        kernel_param_names = []
        data_struct_init = []
        data_struct_names = []

        for key, data_adapter in self._var_data_dict.items():
            clmemtype = 'global'

            cl_data = data_adapter.get_opencl_data()

            if self._check_array_fits_constant_buffer(cl_data, data_adapter.get_opencl_numpy_type()):
                if constant_args_counter < self._max_constant_args:
                    clmemtype = 'constant'
                    constant_args_counter += 1

            param_name = 'var_data_' + str(key)
            data_type = data_adapter.get_data_type().raw_data_type

            if data_adapter.get_data_type().is_vector_type:
                data_type += data_adapter.get_data_type().vector_length

            kernel_param_names.append(clmemtype + ' ' + data_type + '* ' + param_name)

            mult = cl_data.shape[1] if len(cl_data.shape) > 1 else 1
            if len(cl_data.shape) == 1 or cl_data.shape[1] == 1:
                data_struct_names.append(data_type + ' ' + param_name)
                data_struct_init.append(param_name + '[gid * ' + str(mult) + ']')
            else:
                data_struct_names.append(clmemtype + ' ' + data_type + '* ' + param_name)
                data_struct_init.append(param_name + ' + gid * ' + str(mult))

        for key, data_adapter in self._protocol_data_dict.items():
            clmemtype = 'global'

            cl_data = data_adapter.get_opencl_data()

            if self._check_array_fits_constant_buffer(cl_data, data_adapter.get_opencl_numpy_type()):
                if constant_args_counter < self._max_constant_args:
                    clmemtype = 'constant'
                    constant_args_counter += 1

            param_name = 'protocol_data_' + str(key)
            data_type = data_adapter.get_data_type().raw_data_type

            if data_adapter.get_data_type().is_vector_type:
                data_type += str(data_adapter.get_data_type().vector_length)

            kernel_param_names.append(clmemtype + ' ' + data_type + '* ' + param_name)
            data_struct_init.append(param_name)
            data_struct_names.append(clmemtype + ' ' + data_type + '* ' + param_name)

        for key, data_adapter in self._model_data_dict.items():
            clmemtype = 'global'
            param_name = 'model_data_' + str(key)
            data_type = data_adapter.get_data_type().raw_data_type

            if data_adapter.get_data_type().is_vector_type:
                data_type += data_adapter.get_data_type().vector_length

            data_struct_init.append(param_name)

            if isinstance(data_adapter.get_opencl_data(), np.ndarray):
                kernel_param_names.append(clmemtype + ' ' + data_type + '* ' + param_name)
                data_struct_names.append(clmemtype + ' ' + data_type + '* ' + param_name)
            else:
                kernel_param_names.append(data_type + ' ' + param_name)
                data_struct_names.append(data_type + ' ' + param_name)

        data_struct = '''
            typedef struct{
                ''' + ('' if data_struct_names else 'constant void* place_holder;') + '''
                ''' + " ".join((name + ";\n" for name in data_struct_names)) + '''
            } optimize_data;
        '''

        return {'kernel_param_names': kernel_param_names,
                'data_struct_names': data_struct_names,
                'data_struct_init': data_struct_init,
                'data_struct': data_struct}

    def _check_array_fits_constant_buffer(self, array, dtype):
        """Check if the given array when casted to the given type can be fit into the given max_size

        Args:
            array (ndarray): the array we want to fit
            dtype (np data type): the numpy data type we want to use

        Returns:
            boolean: if it fits in the constant memory buffer or not
        """
        return np.product(array.shape) * np.dtype(dtype).itemsize < self._max_constant_buffer_size


def is_scalar(value):
    """Test if the given value is a scalar.

    This function also works with memory mapped array values, in contrast to the numpy is_scalar method.

    Args:
        value: the value to test for being a scalar value

    Returns:
        boolean: if the given value is a scalar or not
    """
    return np.isscalar(value) or (isinstance(value, np.ndarray) and (len(np.squeeze(value).shape) == 0))



import os

import numpy as np
from pkg_resources import resource_filename
from mot.cl_data_type import CLDataType
from mot.model_building.cl_functions.base import ModelFunction
from mot.model_building.cl_functions.parameters import FreeParameter
from mot.model_building.parameter_functions.proposals import GaussianProposal
from mot.model_building.parameter_functions.transformations import ClampTransform, CosSqrClampTransform

__author__ = 'Robbert Harms'
__date__ = "2016-10-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Scalar(ModelFunction):

    def __init__(self, name='Scalar', value=0.0, lower_bound=0.0, upper_bound=float('inf')):
        """A Scalar model function to be used during optimization.

        Args:
            name (str): The name of the model
            value (number or ndarray): The initial value for the single free parameter of this function.
            lower_bound (number or ndarray): The initial lower bound for the single free parameter of this function.
            upper_bound (number or ndarray): The initial upper bound for the single free parameter of this function.
        """
        super(Scalar, self).__init__(
            name,
            'cmScalar',
            (FreeParameter(CLDataType.from_string('mot_float_type'), 's', False, value, lower_bound, upper_bound,
                           parameter_transform=ClampTransform(),
                           sampling_proposal=GaussianProposal(1.0)),))

    def get_cl_header(self):
        """See base class for details"""
        path = resource_filename('mot', 'data/opencl/model_functions/Scalar.h')
        return self._get_cl_dependency_headers() + "\n" + open(os.path.abspath(path), 'r').read()

    def get_cl_code(self):
        """See base class for details"""
        path = resource_filename('mot', 'data/opencl/model_functions/Scalar.cl')
        return self._get_cl_dependency_code() + "\n" + open(os.path.abspath(path), 'r').read()


class Weight(Scalar):

    def __init__(self, name='Weight', value=0.5, lower_bound=0.0, upper_bound=1.0):
        """Implements Scalar model function to add the semantics of representing a Weight.

        Some of the code checks for type Weight, be sure to use this model function if you want to represent a Weight.

        A weight is meant to be a model fraction.

        Args:
            name (str): The name of the model
            value (number or ndarray): The initial value for the single free parameter of this function.
            lower_bound (number or ndarray): The initial lower bound for the single free parameter of this function.
            upper_bound (number or ndarray): The initial upper bound for the single free parameter of this function.
        """
        super(Weight, self).__init__(name=name, value=value, lower_bound=lower_bound, upper_bound=upper_bound)
        self.parameter_list[0].name = 'w'
        self.parameter_list[0].parameter_transform = CosSqrClampTransform()
        self.parameter_list[0].sampling_proposal = GaussianProposal(0.01)
        self.parameter_list[0].perturbation_function = lambda v: np.clip(
            v + np.random.normal(scale=0.05, size=v.shape), 0, 1)



from collections import MutableMapping

import pyopencl as cl
import numpy as np
from ...utils import ParameterCLCodeGenerator, get_float_type_def
from ...cl_routines.base import CLRoutine
from ...load_balance_strategies import Worker


__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ObjectiveListCalculator(CLRoutine):

    def __init__(self, **kwargs):
        """Calculate the objective list, that is it can compute the get_objective_list_function of the given model.

        This evaluates the model and compares it to the problem data to get objective values. This returns
        objective values per observation, for a complete objective function summarized over the observations use
        the :class:`~.objective_calculator.ObjectiveCalculator`.

        This returns a value per problem instance per data point.
        """
        super(ObjectiveListCalculator, self).__init__(**kwargs)

    def calculate(self, model, parameters):
        """Calculate and return the objective lists.

        Args:
            model (AbstractModel): The model to calculate the residuals of.
            parameters (dict or ndarray): The parameters to use in the evaluation of the model
                If a dict is given we assume it is with values for a set of parameters
                If an ndarray is given we assume that we have data for all parameters.

        Returns:
            Return per voxel the objective value (application of the function: "noise_model(eval - data)")
                per protocol item.
        """
        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64

        nmr_inst_per_problem = model.get_nmr_inst_per_problem()
        nmr_problems = model.get_nmr_problems()

        if isinstance(parameters, MutableMapping):
            parameters = np.require(model.get_initial_parameters(parameters), np_dtype,
                                    requirements=['C', 'A', 'O'])
        else:
            parameters = np.require(parameters, np_dtype, requirements=['C', 'A', 'O'])

        objectives = np.zeros((nmr_problems, nmr_inst_per_problem), dtype=np_dtype, order='C')

        workers = self._create_workers(lambda cl_environment: _ObjectiveListCalculatorWorker(
            cl_environment, self.get_compile_flags_list(), model, parameters, objectives))
        self.load_balancer.process(workers, model.get_nmr_problems())

        return objectives


class _ObjectiveListCalculatorWorker(Worker):

    def __init__(self, cl_environment, compile_flags, model, parameters, objectives):
        super(_ObjectiveListCalculatorWorker, self).__init__(cl_environment)

        self._model = model
        self._double_precision = model.double_precision
        self._objectives = objectives
        self._parameters = parameters

        self._var_data_dict = model.get_problems_var_data()
        self._protocol_data_dict = model.get_problems_protocol_data()
        self._model_data_dict = model.get_model_data()

        self._all_buffers, self._residuals_buffer = self._create_buffers()
        self._kernel = self._build_kernel(compile_flags)

    def __del__(self):
        list(buffer.release() for buffer in self._all_buffers)

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start
        event = self._kernel.get_objectives(self._cl_run_context.queue, (int(nmr_problems), ), None, *self._all_buffers,
                                            global_offset=(int(range_start),))
        return [self._enqueue_readout(self._residuals_buffer, self._objectives, range_start, range_end, [event])]

    def _create_buffers(self):
        objectives_buffer = cl.Buffer(self._cl_run_context.context,
                                  cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR,
                                  hostbuf=self._objectives)

        all_buffers = [cl.Buffer(self._cl_run_context.context,
                                 cl.mem_flags.READ_ONLY | cl.mem_flags.USE_HOST_PTR,
                                 hostbuf=self._parameters),
                       objectives_buffer]

        for data in self._var_data_dict.values():
            all_buffers.append(cl.Buffer(self._cl_run_context.context,
                                         cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                                         hostbuf=data.get_opencl_data()))

        constant_buffers = self._generate_constant_buffers(self._protocol_data_dict, self._model_data_dict)
        all_buffers.extend(constant_buffers)

        return all_buffers, objectives_buffer

    def _get_kernel_source(self):
        nmr_inst_per_problem = self._model.get_nmr_inst_per_problem()
        nmr_params = self._parameters.shape[1]

        param_code_gen = ParameterCLCodeGenerator(self._cl_environment.device, self._var_data_dict,
                                                  self._protocol_data_dict, self._model_data_dict)

        kernel_param_names = ['global mot_float_type* params', 'global mot_float_type* objectives']
        kernel_param_names.extend(param_code_gen.get_kernel_param_names())

        kernel_source = '''
            #define NMR_INST_PER_PROBLEM ''' + str(nmr_inst_per_problem) + '''
        '''

        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += param_code_gen.get_data_struct()
        kernel_source += self._model.get_objective_list_function('calculateObjectiveList')
        kernel_source += '''
            __kernel void get_objectives(
                ''' + ",\n".join(kernel_param_names) + '''
                ){
                    int gid = get_global_id(0);
                    ''' + param_code_gen.get_data_struct_init_assignment('data') + '''

                    mot_float_type x[''' + str(nmr_params) + '''];
                    for(int i = 0; i < ''' + str(nmr_params) + '''; i++){
                        x[i] = params[gid * ''' + str(nmr_params) + ''' + i];
                    }

                    mot_float_type tmp_results[NMR_INST_PER_PROBLEM];
                    calculateObjectiveList(&data, x, tmp_results);

                    global mot_float_type* result = objectives + gid * NMR_INST_PER_PROBLEM;
                    for(int i = 0; i < NMR_INST_PER_PROBLEM; i++){
                        result[i] = tmp_results[i];
                    }
            }
        '''
        return kernel_source



from mot.random123 import get_random123_cl_code, RandomStartingPoint
from .base import AbstractParallelOptimizer, AbstractParallelOptimizerWorker

__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class SimulatedAnnealing(AbstractParallelOptimizer):

    default_patience = 500

    def __init__(self, **kwargs):
        """Use Simulated Annealing to calculate the optimum.

        This does not use the parameter codec, even if set to True. This because the priors (should) already
        cover the bounds.

        This implementation uses an adapted Metropolis Hasting algorithm to find the optimum, hence, this requires
        that the model is a implementation of SampleModelInterface.

        Args:
            patience (int):
                Used to set the maximum number of samples to patience*(number_of_parameters+1)
            optimizer_settings (dict): the optimization options. Contains:
                - proposal_update_intervals (int): the interval by which we update the proposal std.

        """
        kwargs['patience'] = kwargs.get('patience', self.default_patience) or self.default_patience
        kwargs['use_param_codec'] = False
        super(SimulatedAnnealing, self).__init__(**kwargs)
        kwargs['optimizer_options'] = kwargs.get('optimizer_options', {}) or {}

        self.proposal_update_intervals = kwargs['optimizer_options'].get('proposal_update_intervals', 50) or 50
        self._annealing_schedule = ExponentialCoolingSchedule()
        self._initial_temperature_strategy = SimpleInitialTemperatureStrategy()

    def _get_worker_generator(self, *args):
        return lambda cl_environment: SimulatedAnnealingWorker(cl_environment, *args, patience=self.patience,
                                                               proposal_update_intervals=self.proposal_update_intervals,
                                                               annealing_schedule=self._annealing_schedule,
                                                               init_temp_strategy=self._initial_temperature_strategy)


class SimulatedAnnealingWorker(AbstractParallelOptimizerWorker):

    def __init__(self, *args, **kwargs):
        self.patience = kwargs.pop('patience')
        self.proposal_update_intervals = kwargs.pop('proposal_update_intervals')
        self._annealing_schedule = kwargs.pop('annealing_schedule')
        self._init_temp_strategy = kwargs.pop('init_temp_strategy')

        self._rand123_starting_point = RandomStartingPoint()

        super(SimulatedAnnealingWorker, self).__init__(*args, **kwargs)
        self._use_param_codec = False

    def _get_optimizer_cl_code(self):
        kernel_source = self._get_evaluate_function()

        kernel_source += get_random123_cl_code()

        kernel_source += self._model.get_log_prior_function('getLogPrior')
        kernel_source += self._model.get_proposal_function('getProposal')
        kernel_source += self._model.get_proposal_state_update_function('updateProposalState')

        if not self._model.is_proposal_symmetric():
            kernel_source += self._model.get_proposal_logpdf('getProposalLogPDF')

        kernel_source += self._model.get_log_likelihood_function('getLogLikelihood', full_likelihood=False)

        kernel_source += self._get_sampling_code()
        return kernel_source

    def _get_optimizer_call_name(self):
        return 'simulated_annealing'

    def _get_sampling_code(self):
        proposal_state_size = len(self._model.get_proposal_state())
        proposal_state = '{' + ', '.join(map(str, self._model.get_proposal_state())) + '}'
        acceptance_counters_between_proposal_updates = '{' + ', '.join('0' * self._nmr_params) + '}'

        kernel_source = self._annealing_schedule.get_temperature_update_cl_function()
        kernel_source += '''
            void _update_proposals(mot_float_type* const proposal_state, uint* const ac_between_proposal_updates,
                                   uint* const proposal_update_count){

                *proposal_update_count += 1;

                if(*proposal_update_count == ''' + str(self.proposal_update_intervals) + '''){
                    updateProposalState(ac_between_proposal_updates,
                                        ''' + str(self.proposal_update_intervals) + ''',
                                        proposal_state);

                    for(int i = 0; i < ''' + str(proposal_state_size) + '''; i++){
                        ac_between_proposal_updates[i] = 0;
                    }

                    *proposal_update_count = 0;
                }
            }

            void _update_state(mot_float_type* const x,
                               void* rng_data,
                               double* const current_likelihood,
                               mot_float_type* const current_prior,
                               const optimize_data* const data,
                               mot_float_type* const proposal_state,
                               uint* const ac_between_proposal_updates,
                               mot_float_type* temperature){

                mot_float_type new_prior;
                double new_likelihood;
                double bayesian_f;
                mot_float_type old_x;

                #pragma unroll 1
                for(int k = 0; k < ''' + str(self._nmr_params) + '''; k++){

                    old_x = x[k];
                    x[k] = getProposal(k, x[k], rng_data, proposal_state);

                    new_prior = getLogPrior(x);

                    if(exp(new_prior) > 0){
                        new_likelihood = getLogLikelihood(data, x);
        '''
        if self._model.is_proposal_symmetric():
            kernel_source += '''
                        bayesian_f = exp(((new_likelihood + new_prior) -
                                            (*current_likelihood + *current_prior)) / *temperature);
                '''
        else:
            kernel_source += '''
                        mot_float_type x_to_prop = getProposalLogPDF(k, old_x, x[k], proposal_state);
                        mot_float_type prop_to_x = getProposalLogPDF(k, x[k], x[k], proposal_state);

                        bayesian_f = exp(((new_likelihood + new_prior + x_to_prop) -
                                          (*current_likelihood + *current_prior + prop_to_x)) / *temperature);
                '''
        kernel_source += '''
                        if(new_likelihood > *current_likelihood || frand(rng_data) < bayesian_f){
                            *current_likelihood = new_likelihood;
                            *current_prior = new_prior;
                            ac_between_proposal_updates[k]++;
                        }
                        else{
                            x[k] = old_x;
                        }
                    }
                    else{
                        x[k] = old_x;
                    }
                }
            }
        '''
        kernel_source += self._init_temp_strategy.get_init_temp_cl_function()
        kernel_source += '''
            int simulated_annealing(mot_float_type* const x, const void* const data){
                rand123_data rand123_rng_data = ''' + self._get_rand123_init_cl_code() + ''';
                void* rng_data = (void*)&rand123_rng_data;

                uint proposal_update_count = 0;

                mot_float_type proposal_state[] = ''' + proposal_state + ''';
                uint ac_between_proposal_updates[] = ''' + acceptance_counters_between_proposal_updates + ''';

                double current_likelihood = getLogLikelihood((optimize_data*)data, x);
                mot_float_type current_prior = getLogPrior(x);

                mot_float_type temperature;
                mot_float_type min_temp;
                mot_float_type initial_temp;
                its_get_initial_temperature(&temperature, &min_temp, &initial_temp,
                                            x, rng_data, &current_likelihood, &current_prior,
                                            data, proposal_state, ac_between_proposal_updates,
                                            &proposal_update_count);

                for(uint step = 0; step < ''' + str(self.patience * (self._nmr_params + 1)) + '''; step++){
                    as_update_temperature(&temperature, min_temp, initial_temp, step,
                                          (uint) ''' + str(self.patience * (self._nmr_params + 1)) + ''');

                    if(temperature <= 0.0f){
                        return 2;
                    }

                    _update_state(x, rng_data, &current_likelihood, &current_prior,
                                  data, proposal_state, ac_between_proposal_updates, &temperature);
                    _update_proposals(proposal_state, ac_between_proposal_updates, &proposal_update_count);
                }

                return 1;
            }
        '''
        return kernel_source

    def _get_rand123_init_cl_code(self):
        key = self._rand123_starting_point.get_key()
        counter = self._rand123_starting_point.get_counter()

        if len(key):
            return 'rand123_initialize_data_extra_precision((uint[]){%(c0)r, %(c1)r, %(c2)r, %(c3)r}, ' \
                   '(uint[]){%(k0)r, %(k1)r})' % {'c0': counter[0], 'c1': counter[1],
                                                  'c2': counter[2], 'c3': counter[3],
                                                  'k0': key[0], 'k1': counter[1]}
        else:
            return 'rand123_initialize_data((uint[]){%(c0)r, %(c1)r, %(c2)r, %(c3)r})' \
                   % {'c0': counter[0], 'c1': counter[1], 'c2': counter[2], 'c3': counter[3]}


class InitialTemperatureStrategy(object):

    def __init__(self, *args, **kwargs):
        """The base class for strategies for determining the initial temperature to use during annealing.
        """

    def get_init_temp_cl_function(self):
        """Get the function called by the annealing routine to determine the initial temperature.
        """


class SimpleInitialTemperatureStrategy(InitialTemperatureStrategy):

    def __init__(self, min_temp=0, initial_temp=1e5, **kwargs):
        """Sets the initial temperature to a predefined value.
        """
        super(SimpleInitialTemperatureStrategy, self).__init__(**kwargs)
        self.min_temp = min_temp
        self.initial_temp = initial_temp

    def get_init_temp_cl_function(self):
        return '''
            void its_get_initial_temperature(
                    mot_float_type* temperature,
                    mot_float_type* min_temp,
                    mot_float_type* initial_temp,
                    mot_float_type* const x,
                    void* rng_data,
                    double* const current_likelihood,
                    mot_float_type* const current_prior,
                    const optimize_data* const data,
                    mot_float_type* const proposal_state,
                    uint* const ac_between_proposal_updates,
                    uint* const proposal_update_count){

                *min_temp = ''' + str(self.min_temp) + ''';
                *initial_temp = ''' + str(self.initial_temp) + ''';
                *temperature = *initial_temp;
            }
        '''


class AnnealingSchedule(object):

    def __init__(self, *args, **kwargs):
        """The base class for the annealing schedule.

        Implementing classes can implement an annealing schedule to update the temperature during annealing.
        """

    def get_temperature_update_cl_function(self):
        """Get the function called by the annealing routine to update the temperature.

        This must return a CL string with a function with the following signature
        (where the prefix 'as' stands for Annealing Schedule):

        .. code-block:: c

            void as_update_temperature(mot_float_type* temperature, const mot_float_type min_temp,
                                       const mot_float_type initial_temp, const uint step, const uint nmr_steps);

        Here ``temperature`` is the value to control, ``min_temp`` is the minimum temperature allowed,
        ``initial_temp`` was the initial temperature, ``step`` is the current step and ``nmr_steps``
        is the maximum number of steps.

        Returns:
            str: the temperature update function
        """


class ExponentialCoolingSchedule(AnnealingSchedule):

    def __init__(self, damping_factor=0.95, **kwargs):
        """A simple exponential cooling schedule.

        The temperature ``T`` at time ``k+1`` is given by:

        .. code-block:: c

            T_k+1 = a * T_k

        Where ``0 < a < 1`` is the damping factor.
        """
        self.damping_factor = damping_factor
        super(ExponentialCoolingSchedule, self).__init__(**kwargs)

    def get_temperature_update_cl_function(self):
        return '''
            void as_update_temperature(mot_float_type* temperature, const mot_float_type min_temp,
                                       const mot_float_type initial_temp, const uint step, const uint nmr_steps){
                *temperature *= ''' + (str(self.damping_factor)) + ''';

                 if(*temperature < min_temp){
                    *temperature = min_temp;
                 }
            }
        '''



import pyopencl as cl
import numpy as np
from ...utils import ParameterCLCodeGenerator, get_float_type_def
from ...cl_routines.base import CLRoutine
from ...load_balance_strategies import Worker


__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class FinalParametersTransformer(CLRoutine):

    def __init__(self, cl_environments=None, load_balancer=None):
        """CL code for applying the final parameter transforms.

        Some of the models may contain parameter dependencies. These dependencies may have side-effects and change
        multiple parameters. This change occurs right before the parameters are entered in the model evaluation function
        right before the actual model is evaluated.

        Suppose an optimization routine finds a set of parameters X to the the optimal set of parameters. In the
        evaluation function this set of parameters might have been transformed to a new set of parameters X' by the
        parameter dependencies. Since we, in the end, are interested in the set of parameters X', we have to apply
        the exact same transformations at the end of the optimization routine as happened in the evaluation function.

        This class supports running those transformations.
        """
        super(FinalParametersTransformer, self).__init__(cl_environments, load_balancer)

    def transform(self, model, parameters):
        """This transforms the parameters matrix in place. Using the final parameters transforms.

        Args:
            model (OptimizeModelInterface): the model for which we want to do the final parameter transformations
            parameters (ndarray): the parameters we want to transform
        """
        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64

        parameters = np.require(parameters, np_dtype, requirements=['C', 'A', 'O', 'W'])
        var_data_dict = model.get_problems_var_data()
        protocol_data_dict = model.get_problems_protocol_data()
        model_data_dict = model.get_model_data()

        if model.get_final_parameter_transformations():
            workers = self._create_workers(lambda cl_environment: _FPTWorker(
                cl_environment, self.get_compile_flags_list(), model, parameters,
                var_data_dict, protocol_data_dict, model_data_dict))
            self.load_balancer.process(workers, model.get_nmr_problems())

        return parameters


class _FPTWorker(Worker):

    def __init__(self, cl_environment, compile_flags, model, parameters, var_data_dict, protocol_data_dict,
                 model_data_dict):
        super(_FPTWorker, self).__init__(cl_environment)

        self._parameters = parameters
        self._nmr_params = parameters.shape[1]
        self._model = model
        self._var_data_dict = var_data_dict
        self._protocol_data_dict = protocol_data_dict
        self._model_data_dict = model_data_dict
        self._double_precision = model.double_precision
        self._all_buffers, self._parameters_buffer = self._create_buffers()
        self._kernel = self._build_kernel(compile_flags)

    def __del__(self):
        list(buffer.release() for buffer in self._all_buffers)

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start

        event = self._kernel.transform(self._cl_run_context.queue, (int(nmr_problems), ), None, *self._all_buffers,
                                       global_offset=(int(range_start),))
        return [self._enqueue_readout(self._parameters_buffer, self._parameters, range_start, range_end, [event])]

    def _create_buffers(self):
        all_buffers = []
        parameters_buffer = cl.Buffer(self._cl_run_context.context,
                                      cl.mem_flags.READ_WRITE | cl.mem_flags.USE_HOST_PTR,
                                      hostbuf=self._parameters)
        all_buffers.append(parameters_buffer)

        for data in self._var_data_dict.values():
            all_buffers.append(cl.Buffer(self._cl_run_context.context,
                                         cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                                         hostbuf=data.get_opencl_data()))

        constant_buffers = self._generate_constant_buffers(self._protocol_data_dict, self._model_data_dict)
        all_buffers.extend(constant_buffers)
        return all_buffers, parameters_buffer

    def _get_kernel_source(self):
        param_code_gen = ParameterCLCodeGenerator(self._cl_environment.device,
                                                  self._var_data_dict, self._protocol_data_dict, self._model_data_dict)

        kernel_param_names = ['global mot_float_type* params']
        kernel_param_names.extend(param_code_gen.get_kernel_param_names())

        kernel_source = ''
        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += param_code_gen.get_data_struct()
        kernel_source += self._model.get_final_parameter_transformations('applyFinalParameterTransformations')
        kernel_source += '''
            __kernel void transform(
                ''' + ",\n".join(kernel_param_names) + '''
                ){
                    int gid = get_global_id(0);
                    ''' + param_code_gen.get_data_struct_init_assignment('data') + '''

                    mot_float_type x[''' + str(self._nmr_params) + '''];
                    for(int i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                        x[i] = params[gid * ''' + str(self._nmr_params) + ''' + i];
                    }

                    applyFinalParameterTransformations(&data, x);

                    for(int i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                        params[gid * ''' + str(self._nmr_params) + ''' + i] = x[i];
                    }
            }
        '''
        return kernel_source



import os
from .parameters import FreeParameter

__author__ = 'Robbert Harms'
__date__ = "2016-10-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CLFunction(object):

    def __init__(self, return_type, function_name, parameter_list):
        """The header to one of the CL library functions.

        Ideally the functions map bijective to the CL function files.

        Args:
            return_type (str): Return type of the CL function.
            function_name (string): The name of the CL function
            parameter_list (list of mot.model_building.cl_functions.parameters.CLFunctionParameter): The list of parameters required for this function
        """
        super(CLFunction, self).__init__()
        self._return_type = return_type
        self._function_name = function_name
        self._parameter_list = parameter_list

    @property
    def return_type(self):
        """Get the type (in CL naming) of the returned value from this function.

        Returns:
            str: The return type of this CL function. (Examples: double, int, double4, ...)
        """
        return self._return_type

    @property
    def cl_function_name(self):
        """Return the name of the implemented CL function

        Returns:
            str: The name of this CL function
        """
        return self._function_name

    @property
    def parameter_list(self):
        """Return the list of parameters from this CL function.

        Returns:
            A list containing instances of CLFunctionParameter."""
        return self._parameter_list

    def get_cl_header(self):
        """Get the CL header for this function and all its dependencies

        Returns:
            str: The CL header code for inclusion in CL source code.
        """
        return ''

    def get_cl_code(self):
        """Get the function code for this function and all its dependencies.

        Returns:
            str: The CL header code for inclusion in CL source code.
        """
        return ''

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        return type(self) == type(other)

    def __ne__(self, other):
        return type(self) != type(other)


class DependentCLFunction(CLFunction):

    def __init__(self, return_type, function_name, parameter_list, dependency_list):
        """A CL function with dependencies on multiple other CLFunctions.

        Args:
            return_type (str): Return type of the CL function.
            function_name (string): The name of the CL function
            parameter_list (list of mot.model_building.cl_functions.parameters.CLFunctionParameter): The list of parameters required for this function
            dependency_list (list of CLFunction): The list of CLFunctions this function is dependent on
        """
        super(DependentCLFunction, self).__init__(return_type, function_name, parameter_list)
        self._dependency_list = dependency_list

    def _get_cl_dependency_headers(self):
        """Get the CL code for all the headers for all the dependencies.

        Returns:
            str: The CL code with the headers.
        """
        header = ''
        for d in self._dependency_list:
            header += d.get_cl_header() + "\n"
        return header

    def _get_cl_dependency_code(self):
        """Get the CL code for all the CL code for all the dependencies.

        Returns:
            str: The CL code with the actual code.
        """
        code = ''
        for d in self._dependency_list:
            code += d.get_cl_code() + "\n"
        return code


class ModelFunction(DependentCLFunction):

    def __init__(self, name, cl_function_name, parameter_list, dependency_list=()):
        """This CL function is for all estimable models

        Args:
            name (str): The name of the model
            cl_function_name (string): The name of the CL function
            parameter_list (list or tuple of CLFunctionParameter): The list of parameters required for this function
            dependency_list (list or tuple of CLFunction): The list of CLFunctions this function is dependent on
        """
        super(ModelFunction, self).__init__('double', cl_function_name, parameter_list, dependency_list)
        self._name = name

    @property
    def name(self):
        """Get the name of this model function.

        Returns:
            str: The name of this model function.
        """
        return self._name

    def get_free_parameters(self):
        """Get all parameters whose state instance is one of the type CLPSFreeState.

        Returns:
            A list of parameters whose type matches the CLPSFreeState state type.
        """
        return self.get_parameters_of_type(FreeParameter)

    def get_parameters_of_type(self, instance_types):
        """Get all parameters whose state instance is one of the given types.

        Args:
            instance_types (list of DataType class names, or single DataType classname);
                The instance type we want to get all the parameters of.

        Returns:
            A list of parameters whose type matches one or more of the given types.
        """
        return list([p for p in self.parameter_list if isinstance(p, instance_types)])

    def init(self, param_name, value):
        """Set the parameter (by the param_name) to the given value (single value or array)

        This generally only works for parameters with the state CLPSFreeState, for all other states it may fail.

        Args:
            param_name (str): the name of the parameter to initialize
            value (number or ndarray): the optional value to initialize it to
        """
        self.get_parameter_by_name(param_name).value = value
        return self

    def unfix(self, param_name):
        """Unfix this parameter, which means, make sure it is no longer fixed if it was fixed.

        This generally only works for parameters with the state CLPSFreeState, for all other states it may fail.

        Args:
           param_name (str): the name of the parameter
        """
        self.get_parameter_by_name(param_name).unfix()
        return self

    def fix(self, param_name, value=None):
        """Fix this parameter to the given value (single value or array of values)

        This generally only works for parameters with the state CLPSFreeState, for all other states it may fail.

        Args:
            param_name (str): the name of the parameter
            value (number or ndarray): the value to fix it to (one value, or an array with one value per voxel)
                If value is None, the value is fixed to the default value (if possible)
        """
        if value is not None:
            self.get_parameter_by_name(param_name).fix_to(value)
        self.get_parameter_by_name(param_name).fixed = True
        return self

    def ub(self, param_name, value):
        """Set the upper bound for this parameter to the specific value.

        This may be used during parameter transformations and or sampling proposal generation and or the sampling
        prior function.

        This generally only works for parameters with the state CLPSFreeState, for all other states it may fail.

        Args:
            param_name: the name of the parameter
            value: the upper bound
        """
        self.get_parameter_by_name(param_name).upper_bound = value
        return self

    def lb(self, param_name, value):
        """Set the lower bound for this parameter to the specific value.

        This may be used during parameter transformations and or sampling proposal generation and or the sampling
        prior function.

        This generally only works for parameters with the state CLPSFreeState, for all other states it may fail.

        Args:
            param_name: the name of the parameter
            value: the lower bound
        """
        self.get_parameter_by_name(param_name).lower_bound = value
        return self

    def has_parameter_by_name(self, param_name):
        """Check if this model function has a parameter with the given name.

        Args:
            param_name (str): The name of the parameter to check for existence.

        Returns:
            boolean: True if a parameter of the given name is attached to this function, else otherwise.
        """
        try:
            self.get_parameter_by_name(param_name)
            return True
        except KeyError:
            return False

    def p(self, param_name):
        """Get a parameter by name.

        This function is a short for the function get_parameter_by_name()

        Args:
            param_name (str): The name of the parameter to return

        Returns:
            ClFunctionParameter: the parameter of the given name
        """
        return self.get_parameter_by_name(param_name)

    def get_parameter_by_name(self, param_name):
        """Get a parameter by name.

        Args:
            param_name (str): The name of the parameter to return

        Returns:
            ClFunctionParameter: the parameter of the given name

        Raises:
            KeyError: if the parameter could not be found.
        """
        for e in self.parameter_list:
            if e.name == param_name:
                return e
        raise KeyError('The parameter with the name "{}" could not be found.'.format(param_name))

    def get_extra_results_maps(self, results_dict):
        """Get extra results maps with extra output from this model function.

        This is used by the function finalize_optimization_results() from the ModelBuilder to add extra maps
        to the resulting dictionary.

        Suppose a model has a parameter that can be viewed in multiple ways. It would be nice to be able
        to output maps for that parameter in multiple ways such that the amount of post-processing is as least as
        possible.

        For example, suppose a model calculates an angle (theta) and a radius (r). Perhaps we would like to return
        the cartesian coordinate of that point alongside the polar coordinates. This function allows you (indirectly)
        to add the additional maps.

        Do not modify the dictionary in place.

        Args:
            results_dict (dict): The result dictionary with all the maps you need and perhaps other maps from other
                models as well. The maps are 1 dimensional, a long list of values for all the voxels in the ROI.

        Returns:
            dict: A new dictionary with the additional maps to add.
        """
        return {}

    def get_cl_dependency_headers(self):
        """Get the CL code for all the headers for all the dependencies.

        Returns:
            str: The CL code with the headers.
        """
        return self._get_cl_dependency_headers()

    def get_cl_dependency_code(self):
        """Get the CL code for all the CL code for all the dependencies.

        Returns:
            str: The CL code with the actual code.
        """
        return self._get_cl_dependency_code()


class LibraryFunction(DependentCLFunction):

    def __init__(self, return_type, function_name, parameter_list, cl_header_file,
                 cl_code_file, var_replace_dict, dependency_list):
        """Create a CL function for a library function.

        These functions are not meant to be optimized, but can be used a helper functions for the models.

        Args:
            return_type (str): Return type of the CL function.
            function_name (str): The name of the CL function
            cl_header_file (str): The location of the header file .h or .ph
            cl_code_file (str): The location of the code file .c or .pcl
            var_replace_dict (dict): In the cl_header and cl_code file these replacements will be made
                (using the % format function of Python)
            parameter_list (list or tuple of mot.model_building.cl_functions.parameters.CLFunctionParameter): The list of parameters required for this function
            dependency_list (list or tuple of mot.model_building.cl_functions.base.CLFunction): The list of CLFunctions this function is dependent on
        """
        super(LibraryFunction, self).__init__(return_type, function_name, parameter_list, dependency_list)
        self._cl_header_file = cl_header_file
        self._cl_code_file = cl_code_file
        self._var_replace_dict = var_replace_dict

    def get_cl_header(self):
        """Get the CL header for this function and all its dependencies.

        Returns:
            str: The CL code for the header
        """
        header = self._get_cl_dependency_headers()
        header += "\n"
        body = open(os.path.abspath(self._cl_header_file), 'r').read()
        if self._var_replace_dict:
            body = body % self._var_replace_dict
        return header + "\n" + body

    def get_cl_code(self):
        """Get the CL code for this function and all its dependencies.

        Returns:
            str: The CL code for the body of the function
        """
        code = self._get_cl_dependency_code()
        code += "\n"
        body = open(os.path.abspath(self._cl_code_file), 'r').read()
        if self._var_replace_dict:
            body = body % self._var_replace_dict
        return code + "\n" + body



__author__ = 'robbert'
__date__ = "5/15/14"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractCodec(object):

    def get_cl_decode_function(self, fname='decodeParameters'):
        """Get a CL function that can transform the parameters from encoded space to model space.

        The signature of the CL function is:

        .. code-block:: c

            void <fname>(const mot_float_type* x);

        Args:
            fname (str): The CL function name to use

        Returns:
            str: An OpenCL function that is used in the CL kernel to transform the parameters from encoded space to
                model space so they can be used as input to the model.
        """
        pass

    def get_cl_encode_function(self, fname='encodeParameters'):
        """Get a CL function that can transform the parameters from model space to an encoded space.

        The signature of the CL function is:

        .. code-block:: c

            void <fname>(const mot_float_type* x);

        Args:
            fname (str): The CL function name to use

        Returns:
            str: An OpenCL function that is used in the CL kernel to transform the parameters from model space to
                encoded space so they can be used as input to an CL routine.
        """
        pass

    def get_nmr_parameters(self):
        """Get the number of parameters that are encoded and decoded.

        Returns:
            int: the number of parameters that are used in the decoding and encoding.
        """
        pass


class IdentityCodec(AbstractCodec):

    def __init__(self, nmr_parameters):
        """Create an identity codec.

        Input == output for this codec.

        Args:
            nmr_parameters (int): The number of parameters for the codec.
        """
        super(IdentityCodec, self).__init__()
        self._nmr_parameters = nmr_parameters

    def get_cl_decode_function(self, fname='decodeParameters', parameter_offset=0):
        func = '''
            void ''' + fname + '''(mot_float_type* x){'''
        for i in range(self._nmr_parameters):
            func += "\n" + "\t"*4 + 'x[' + str(i + parameter_offset) + '] = x[' + str(i + parameter_offset) + '];'
        return func + '''
            }
        '''

    def get_cl_encode_function(self, fname='encodeParameters', parameter_offset=0):
        func = '''
            void ''' + fname + '''(mot_float_type* x){'''
        for i in range(self._nmr_parameters):
            func += "\n" + "\t"*4 + 'x[' + str(i + parameter_offset) + '] = x[' + str(i + parameter_offset) + '];'
        return func + '''
            }
        '''

    def get_nmr_parameters(self):
        return self._nmr_parameters


class CodecBuilder(AbstractCodec):

    def __init__(self, enc_trans_list, dec_trans_list):
        """Build a codec out of the list of encode and decode transformation functions.

        Args:
            enc_trans_list (tuple): a listing of the encoding transformations, which are called in the given order.
                Each format should be a string with the format option {0} which is meant for the name of the array.
                The following is an example input (with two functions)::

                    ("{0}[0] = sqrt({0}[0]);",
                     "{0}[1] = sqrt({0}[1]);")

            enc_trans_list (tuple): a listing of the decoding transformations, which are called in the given order.
                Each format should be a string with the format option {0} which is meant for the name of the array.
                Example::

                        {0}[0] = {0}[0] * {0}[0];
                        {0}[1] = {0}[1] * {0}[1];
        """
        super(CodecBuilder, self).__init__()
        self._nmr_parameters = len(enc_trans_list)
        self._enc_trans_list = enc_trans_list
        self._dec_trans_list = dec_trans_list

    def get_cl_decode_function(self, fname='decodeParameters', parameter_offset=0):
        func = '''
            void ''' + fname + '''(mot_float_type* x){'''
        for d in self._dec_trans_list:
            func += "\n" + "\t"*4 + d.format('x')
        return func + '''
            }
        '''

    def get_cl_encode_function(self, fname='encodeParameters', parameter_offset=0):
        func = '''
            void ''' + fname + '''(mot_float_type* x){'''
        for d in self._enc_trans_list:
            func += "\n" + "\t"*4 + d.format('x')
        return func + '''
            }
        '''

    def get_nmr_parameters(self):
        return self._nmr_parameters


class CompositeCodec(AbstractCodec):

    def __init__(self, codec_list):
        """Given an ordered tuple of codecs, build a composite codec with the right indices.

        The encode and decode function of this class are the combination of the encode and decode functions of the
        given codecs in the codec list.

        Suppose we have four parameters and two codecs, one for one parameter and one for three parameters. This codec
        will construct a codec function which uses the first codec for the first parameter and the second codec for
        the remaining parameters. If the codecs were reversed, the first three parameters would be handled by the
        first codec and the last by the second. Order of the codec matters.

        Args:
            codec_list (tuple of AbstractCodecs): A list of codecs, used in that order.
        """
        super(CompositeCodec, self).__init__()
        self._codec_list = codec_list

    def get_cl_decode_function(self, fname='decodeParameters'):
        func = ''
        param_offset = 0
        for i, f in enumerate(self._codec_list):
            func += f.get_cl_decode_function(fname=(fname + '_' + str(i)), parameter_offset=param_offset)
            param_offset += f.get_decoded_space_width()

        func += '''
            void ''' + fname + '''(mot_float_type* x){
        '''
        for i in range(len(self._codec_list)):
            func += fname + '_' + str(i) + '(x);' + "\n"
        func += '''
            }
        '''
        return func

    def get_cl_encode_function(self, fname='encodeParameters'):
        func = ''
        param_offset = 0
        for i, f in enumerate(self._codec_list):
            func += f.get_cl_encode_function(fname=(fname + '_' + str(i)), parameter_offset=param_offset)
            param_offset += f.get_decoded_space_width()
        func += '''
            void ''' + fname + '''(mot_float_type* x){
        '''
        for i in range(len(self._codec_list)):
            func += fname + '_' + str(i) + '(x);' + "\n"
        func += '''
            }
        '''
        return func

    def get_nmr_parameters(self):
        return sum([x.get_decoded_space_width() for x in self._codec_list])
